gmsh_map = {
    "Triangle 3": "S3",
    "Triangle 6": "STRI65",
    "Quadrilateral 4": "S4R",
    "Quadrilateral 8": "S8R",
    "Line 2": "B31",
    "Line 3": "B32",
    "Tetrahedron 4": "C3D4",
    "Tetrahedron 10": "C3D10",
}
